#pragma once
#include <string>

#include <glm/glm.hpp>

#include <imgui.h>
#include "../../libs/guizmo/ImGuizmo.h"

#include "../../YoggieEngine/src/YoggieEngine.h"
#include "../../src/Scene/Entity.h"

#include "EditorWindow.h"
#include "EditorConsole.h"
#include "../Project/Project.h"
#include "../AssetManagement/AssetManager.h"

typedef void ( *voidFunction ) (void);
using namespace YoggieEngine;

auto matrix = glm::mat4(1.0f);
auto worldOrigin = glm::mat4(1.0f);


inline void ComponentView(const std::string& componentName, voidFunction func)
{
    ImGuiWindowFlags_ window_flags = ImGuiWindowFlags_None;
    ImGui::PushStyleVar(ImGuiStyleVar_ChildRounding, 5.0f);
    ImGui::BeginChild(componentName.c_str());

    func();

    ImGui::EndChild();
    ImGui::PopStyleVar();
}

class Inspector : EditorWindow {
public:
	Inspector() : EditorWindow("Inspector") {}

	void AddComponentDropDown(Entity& selected ) 
    {
		static char* names[] = { "Script Component", "Camera Component", "Light Component"};
        if (ImGui::Button("Add Component"))
            ImGui::OpenPopup("Component picker");

        ImGui::SameLine();
        if (ImGui::BeginPopup("Component picker")) {

            for (int i = 0; i < IM_ARRAYSIZE(names); i++)
                if (ImGui::MenuItem(names[i])) {
                    std::cout << "Add a " << names[i] << " to "
                        << selected.GetComponent<IdentifierComponent>().name << std::endl;
                }

            ImGui::EndPopup();
        }
        ImGui::NewLine();
	}
	
    void ShowComponents(Entity& selected) 
	{
        auto component = selected.GetComponent<IdentifierComponent>();
        ImGui::InputText("Name:", (char*)component.name.c_str(), component.name.size() * sizeof(char), ImGuiInputTextFlags_ReadOnly);


        if (selected.HasComponent<TransformComponent>()) {
            auto& transform = selected.GetComponent<TransformComponent>();
            if (ImGui::CollapsingHeader("Transform", ImGuiTreeNodeFlags_DefaultOpen)) {
                ImGui::DragFloat3("Position", glm::value_ptr(transform.Position), 0.1f);
                ImGui::DragFloat3("Rotation", glm::value_ptr(transform.Rotation), 0.1f);
                ImGui::DragFloat3("Scale", glm::value_ptr(transform.Scale), 0.1f, 0.0f);
            }
            if (selected.HasComponent<RelationComponent>()) {
                ImGui::Text("Has relation");
            }


        }


        if (selected.HasComponent<Render3DComponent>()) {
            auto& render3d = selected.GetComponent<Render3DComponent>();
            if (ImGui::CollapsingHeader("Render3D", ImGuiTreeNodeFlags_DefaultOpen)) {
                ImGui::ColorEdit3("Colour", glm::value_ptr(render3d.color));
            }
        }
        static bool deferred = true;
        if (selected.HasComponent<LightComponent>()) {
            auto& light = selected.GetComponent<LightComponent>();
            if (ImGui::CollapsingHeader("Light", ImGuiTreeNodeFlags_DefaultOpen)) {
                ImGui::ColorEdit3("Colour", glm::value_ptr(light.Color));
                ImGui::Checkbox("Deferred", &deferred);
            }

        }

        if (selected.HasComponent <CameraComponent>()) {
            auto& camera = selected.GetComponent<CameraComponent>();
            if (ImGui::CollapsingHeader("Camera")) {
                ImGui::DragFloat3("Position:",glm::value_ptr(camera.Position), 0.01f);
                ImGui::DragFloat3("Rotation:", glm::value_ptr(camera.Rotation), 0.01f);

            }
        }

        if (selected.HasComponent<ScriptComponent>()) {
            ComponentView("Scripting", [] {
                ImGui::LabelText("##--", "Hello scripting");
                });
        }
	}

};

class SceneExplorer : EditorWindow {
public:
    SceneExplorer(entt::entity& selected , Scene& scene) : EditorWindow("SceneExplorer") {
        scene.getReg().each([&](entt::entity enttNumber) {
            Entity entity = Entity(enttNumber, &scene);
            auto id = entity.GetComponent<IdentifierComponent>();

            if (ImGui::Selectable(id.name.c_str(), enttNumber == selected)) {
                selected = enttNumber;
            }

        });

    }
};


class Viewport : EditorWindow {
public:
    Viewport (Framebuffer& fb, Camera cam) : EditorWindow("SceneView") 
    {
        ImVec2 WinPos = ImGui::GetWindowPos();
        ImVec2 ContentRegionMin = ImGui::GetWindowContentRegionMin();
        ImVec2 ContentRegionMax = ImGui::GetWindowContentRegionMax();
        ImVec2 ScreenSpaceMin = { ContentRegionMin.x + WinPos.x, ContentRegionMin.y + WinPos.y };
        ImVec2 ScreenSpaceMax = { ContentRegionMax.x + WinPos.x,ContentRegionMax.y + WinPos.y };


        ImGui::Image(
            (void*)(intptr_t)fb.GetColourAttachment(),
            ImVec2{ (float)ImGui::GetWindowWidth(),(float)ImGui::GetWindowHeight() }
        );
        

        ImGuizmo::Enable(true);
        ImGuizmo::SetRect(ScreenSpaceMin.x, ScreenSpaceMin.y,ContentRegionMax.x, ContentRegionMax.y);
        
        glm::mat4 transposed_view = glm::transpose(cam.ViewMatrix);
        
        ImGuizmo::DrawGrid(glm::value_ptr(cam.ViewMatrix), glm::value_ptr(cam.ProjectionMatrix), glm::value_ptr(worldOrigin), 100.0f);
        ImGuizmo::ViewManipulate(glm::value_ptr(transposed_view), 1, { ScreenSpaceMin.x,ScreenSpaceMin.y }, { 90,90 }, 0x22CCCCCC);
       
        // Matrix is the model matrix we would want to manipulate
        //ImGuizmo::Manipulate(glm::value_ptr(cam.ViewMatrix), glm::value_ptr(cam.ProjectionMatrix), ImGuizmo::TRANSLATE, ImGuizmo::WORLD, glm::value_ptr(cam.ViewMatrix));

        if (ImGui::IsWindowFocused() ) 
        {
            isFocused = true;

        }
        else {
            isFocused = false;
        }

    }

    bool isFocused = false;

};


int selectedGfxAPI = 0;
int selectedPhysicsEngine = 0;
glm::vec3 Gravity = glm::vec3(0.0f, -9.81f, 0.0f);
bool ShowAdvancedOptions = false;
bool DebugEngine = false;
class Settings : EditorWindow {
public:
    Settings() : EditorWindow("Settings") 
    {
        ImGui::LabelText("##title-settings", "Fine grain control over the engine!");
        
        if (ImGui::BeginCombo("Graphics API", GraphicsAPI[selectedGfxAPI])) {
            for (int i = 0; i < 3; i++) {
                bool isSelected = i == selectedGfxAPI;
                if (ImGui::Selectable(GraphicsAPI[i], isSelected)) {
                    selectedGfxAPI = i;
                }

                if(isSelected)
                    ImGui::SetItemDefaultFocus();
            }
         
            ImGui::EndCombo();
        }

        ImGui::NewLine();

        if (ImGui::BeginCombo("Physics Engine", PhysicsEngine[selectedPhysicsEngine])) {
            for (int i = 0; i < 2; i++) {
                bool isSelected = i == selectedPhysicsEngine;
                if (ImGui::Selectable(PhysicsEngine[i], isSelected)) {
                    selectedGfxAPI = i;
                }

                if (isSelected)
                    ImGui::SetItemDefaultFocus();
            }

            ImGui::EndCombo();
        }

        ImGui::InputFloat3("Gravity", glm::value_ptr(Gravity));

        ImGui::NewLine();
        if (ImGui::Button("Show Advanced options ")) {
            ShowAdvancedOptions = !ShowAdvancedOptions;
        }

        if (ShowAdvancedOptions) 
        {
            ImGui::Checkbox("Debug Engine", &DebugEngine);

        }

    }

private:


    const char* PhysicsEngine[2] = {
        "PhysX",
        "Jolt Physics"
    };

    const char* GraphicsAPI[3] = {
        "OpenGL",
        "Vulkan",
        "Metal (Apple)"
    };
  
};

class ProjectInfo : EditorWindow {
public:
    ProjectInfo(Project& project) : EditorWindow("Project Info") {
        ImGui::Text("Project: %s", project.GetName().c_str());
        ImGui::Text("Directory: %s", project.GetProjectDirectory().u8string().c_str());
    }
};

class Console : EditorWindow {
public:

Console() : EditorWindow("Console") {
    s_console = std::make_unique<EditorConsole>();
}
void Show() {
    s_console.get()->Draw();
}

private: 
    std::unique_ptr<EditorConsole> s_console;
};

class AssetFinder : EditorWindow {
public:
    AssetFinder() : EditorWindow("Assets") {


        ImGui::DragInt("IconSize", &iconSize, 1, 30, 90);


        if (ImGui::BeginTable("##resources", 3))
        {
            ImGui::PushStyleColor(ImGuiCol_Button, ImVec4(0.f, 0.f, 0.f, 0.f));
            ImGui::PushStyleColor(ImGuiCol_ButtonActive, ImVec4(0.f, 0.f, 0.f, 0.f));
            ImGui::PushStyleColor(ImGuiCol_ButtonHovered, ImVec4(1.f, 1.f, 1.f, 0.2f));

            Texture folderIcon = Texture("rsc/folderIcon.png");
            Texture assetIcon = Texture("rsc/assetIcon.png");


            int row = 0;
            int column = 0;
            for (auto& asset : AssetManager::assets) {
                if (column % 3 == 0) {
                    ImGui::TableNextRow();
                    column = 0;
                    row++;
                }

                ImGui::TableSetColumnIndex(column);

                if (asset.isFolder) {
                    ImGui::ImageButton(
                        (ImTextureID)folderIcon.GetID(),
                        ImVec2{ (float)iconSize,(float)iconSize });
                    ImGui::Text(asset.GetName(), row);

                }
                else {
                    ImGui::ImageButton(
                        (ImTextureID)assetIcon.GetID(),
                        ImVec2{ (float)iconSize, (float)iconSize });
                    ImGui::Text(asset.GetName(), row);

                }


                column++;
            }


            ImGui::PopStyleColor(3);
            ImGui::EndTable();
            const GLuint textures[2]{ assetIcon.GetID(), folderIcon.GetID() };
            glDeleteTextures(2,  textures );

        }

    }
private:
    int iconSize = 60;
};

#define RuntimeControlWindowFlags (ImGuiWindowFlags_)(ImGuiWindowFlags_NoDecoration | ImGuiWindowFlags_NoTitleBar | ImGuiWindowFlags_NoCollapse)

struct ButtonInfo {
    const char* Name;
    ImVec4 Color;
};


class RuntimeControls : EditorWindow {
public:
    RuntimeControls() : EditorWindow("RuntimeControls", RuntimeControlWindowFlags) {
        
        ImGui::SameLine((ImGui::GetWindowContentRegionMax().x / 2) - ( numButtons * buttonSize.x ));
        for (int i = 0; i < numButtons; i++) {
            ImVec4 color = button[i].Color;
            ImGui::PushStyleColor(ImGuiCol_Button, color);
            const float strengthIncrease = 1.5f;
            ImGui::PushStyleColor(
                ImGuiCol_ButtonHovered, 
                ImVec4{
                    color.x * strengthIncrease,
                    color.y * strengthIncrease,
                    color.z * strengthIncrease,
                    color.w 
                } 
            );
            if (ImGui::Button(button[i].Name, buttonSize)) {

            }
            ImGui::PopStyleColor();
            ImGui::PopStyleColor();
            ImGui::SameLine();

        }
    }

private:
    ImVec2 buttonSize = ImVec2{ 90 ,25 };
    unsigned int numButtons = 2;
    ButtonInfo button[2] = {
        {"Play" , ImVec4{ 0.001 * 12 , 0.001 * 201 , 0.001* 69, 1.0f}},
        {"Simulate", ImVec4{ 0.001 * 14, 0.001 * 157, 0.001 * 201, 1.0f}}
    };

};