#include <YoggieEngine.h>
#include <glm/gtc/type_precision.hpp>
#include "Physics.h"

namespace YoggieEngine {

	Physics::Physics()
	{
		// startup PhysX
		mFoundation = PxCreateFoundation(PX_PHYSICS_VERSION, gDefaultAllocatorCallback, gDefaultErrorCallback);
		if (!mFoundation) {
			spdlog::critical("PxCreateFoundation failed!");
		}

		mPvd = physx::PxCreatePvd(*mFoundation);
		if (!mPvd) {
			spdlog::critical("pxCreatPvd failed!");
		}
		transport = PxDefaultPvdSocketTransportCreate(PVD_HOST, 5425, 10);
		mPvd->connect(*transport, PxPvdInstrumentationFlag::eALL);


		mPhysics = PxCreatePhysics(PX_PHYSICS_VERSION, *mFoundation, PxTolerancesScale(), recordMemoryAllocations, mPvd);
		if (!mPhysics) {
			spdlog::critical("pxCreatePhysics failed!");
		}
		gDispatcher = nullptr;
		gMaterial = nullptr;
		mScene = nullptr;


	}

	Physics::~Physics()
	{
		if (mScene)
			mScene->release();

		if (gDispatcher)
			gDispatcher->release();

		if (mPhysics)
			mPhysics->release();

		if (mPvd) {
			if (mPvd->isConnected()) {
				mPvd->disconnect();
			}
			mPvd->release();
		}

		if (transport)
			transport->release();


		if (mFoundation)
			mFoundation->release();
	}

	void Physics::Step(float dt)
	{
		mScene->simulate(1.0f/60.0f);
		mScene->fetchResults(true);
	}

	void Physics::Demo()
	{
		createScene();
		SetupPvdDebug();
		createGroundPlane();
		createStack(PxTransform(PxVec3(0, 0, stackZ -= 10.0f)), 10, 2.0f);
	}

	void Physics::EnableDebugVisuals() {
		mScene->setVisualizationParameter(PxVisualizationParameter::eSCALE, 1.0f);
		mScene->setVisualizationParameter(PxVisualizationParameter::eACTOR_AXES, 2.0f);

	}


	static unsigned int VisualizationVAO, VBO;
	static unsigned int numLines;
	void Physics::SubmitMesh() {
		const PxRenderBuffer& rb = mScene->getRenderBuffer();
		

		std::vector<PxVec3> lines= std::vector<PxVec3>();

		numLines = rb.getNbLines();

		for (unsigned int i = 0; i < numLines; i++) {
			auto line = rb.getLines()[i];
			lines.push_back(line.pos0);
			lines.push_back(line.pos1);

		}

		glGenVertexArrays(1, &VisualizationVAO);
		glGenBuffers(1, &VBO);

		glBindVertexArray(VisualizationVAO);
		glBindBuffer(GL_ARRAY_BUFFER, VBO);
		glBufferData(GL_ARRAY_BUFFER, lines.size() * sizeof(PxVec3), &lines, GL_STATIC_DRAW);

		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);

		glBindVertexArray(0);
		glBindBuffer(GL_ARRAY_BUFFER, 0);


	}


	void Physics::DebugRender(Framebuffer& framebuffer) {
		glDepthMask(GL_FALSE);
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffer.GetId());

		debugDraw.Use();
		glBindVertexArray(VisualizationVAO);
		glDrawArrays(GL_LINES, 0, numLines );
		glBindVertexArray(0);

		glBindFramebuffer(GL_FRAMEBUFFER, 0);
		glDepthMask(GL_TRUE);


	}


	PxRigidDynamic* Physics::createDynamic(const PxTransform& t, const PxGeometry& geometry, const PxVec3& velocity = PxVec3(0)) {
		PxRigidDynamic* dynamic = PxCreateDynamic(*mPhysics, t, geometry, *gMaterial, 10.0f);
		dynamic->setAngularDamping(0.5f);
		dynamic->setLinearVelocity(velocity);
		mScene->addActor(*dynamic);
		return dynamic;
	}

	void Physics::createStack(const PxTransform& t, PxU32 size, PxReal halfextent)
	{
		PxShape* shape = mPhysics->createShape(PxBoxGeometry(halfextent, halfextent, halfextent), *gMaterial);
		for (PxU32 i = 0; i < size; i++) {
			for (PxU32 j = 0; j < size - i; j++) {
				PxTransform localTm(PxVec3(PxReal(j * 2) - PxReal(size - i), PxReal(i * 2 + 1), 0) * halfextent);
				PxRigidDynamic* body = mPhysics->createRigidDynamic(t.transform(localTm));
				body->attachShape(*shape);
				PxRigidBodyExt::updateMassAndInertia(*body, 10.0f);
				mScene->addActor(*body);
			}
		}

		shape->release();

	}

	void Physics::createScene()
	{
		PxSceneDesc sceneDesc(mPhysics->getTolerancesScale());
		sceneDesc.gravity = PxVec3(0.0f, -9.81f, 0.0f);
		gDispatcher = PxDefaultCpuDispatcherCreate(2);
		sceneDesc.cpuDispatcher = gDispatcher;
		sceneDesc.filterShader = PxDefaultSimulationFilterShader;
		mScene = mPhysics->createScene(sceneDesc);

	}

	void Physics::createGroundPlane()
	{
		gMaterial = mPhysics->createMaterial(0.5f, 0.5f, 0.6f);
		PxRigidStatic* groundPlane = PxCreatePlane(*mPhysics, PxPlane(0,1,0,0), *gMaterial);
		mScene->addActor(*groundPlane);

	}

	void Physics::SetupPvdDebug()
	{
		PxPvdSceneClient* pvdClient = mScene->getScenePvdClient();
		if (pvdClient) {
			pvdClient->setScenePvdFlag(PxPvdSceneFlag::eTRANSMIT_CONSTRAINTS, true);
			pvdClient->setScenePvdFlag(PxPvdSceneFlag::eTRANSMIT_CONTACTS, true);
			pvdClient->setScenePvdFlag(PxPvdSceneFlag::eTRANSMIT_SCENEQUERIES, true);
		}


	}

}